@@ -38,7 +38,9 @@
 #define POPT_ARG_MAINCALL	16U+11U	/*!< EXPERIMENTAL: return (*arg) (argc, argv) */
 #define	POPT_ARG_ARGV		12U	/*!< dupe'd arg appended to realloc'd argv array. */
 
-#define POPT_ARG_MASK		0x0000FFFFU
+#define POPT_ARG_MASK		0x000000FFU
+#define POPT_GROUP_MASK		0x0000FF00U
+
 /*@}*/
 
 /** \ingroup popt
@@ -65,7 +67,8 @@
 					/*!< clear arg bit(s) */
 
 #define	POPT_ARGFLAG_SHOW_DEFAULT 0x00800000U /*!< show default value in --help */
-#define	POPT_ARGFLAG_RANDOM	0x00400000U  /*<! random value in [1,arg] */
+#define	POPT_ARGFLAG_RANDOM	0x00400000U  /*!< random value in [1,arg] */
+#define	POPT_ARGFLAG_TOGGLE	0x00200000U  /*!< permit --[no]opt prefix toggle */
 
 /*@}*/
 
@@ -95,6 +98,7 @@
 #define	POPT_ERROR_BADOPERATION	-19	/*!< mutually exclusive logical operations requested */
 #define	POPT_ERROR_NULLARG	-20	/*!< opt->arg should not be NULL */
 #define	POPT_ERROR_MALLOC	-21	/*!< memory allocation failed */
+#define	POPT_ERROR_BADCONFIG	-22	/*!< config file failed sanity test */
 /*@}*/
 
 /** \ingroup popt
@@ -233,6 +237,15 @@ typedef void (*poptCallbackType) (poptContext con,
 	/*@globals internalState @*/
 	/*@modifies internalState @*/;
 
+/** \ingroup popt
+ * Destroy context.
+ * @param con		context
+ * @return		NULL always
+ */
+/*@null@*/
+poptContext poptFreeContext( /*@only@*/ /*@null@*/ poptContext con)
+	/*@modifies con @*/;
+
 /** \ingroup popt
  * Initialize popt context.
  * @param name		context name (usually argv[0] program name)
@@ -251,6 +264,31 @@ poptContext poptGetContext(
 	/*@globals internalState @*/
 	/*@modifies internalState @*/;
 
+/** \ingroup popt
+ * Destroy context (alternative implementation).
+ * @param con		context
+ * @return		NULL always
+ */
+/*@null@*/
+poptContext poptFini( /*@only@*/ /*@null@*/ poptContext con)
+	/*@modifies con @*/;
+
+/** \ingroup popt
+ * Initialize popt context (alternative implementation).
+ * This routine does poptGetContext() and then poptReadConfigFiles().
+ * @param argc		no. of arguments
+ * @param argv		argument array
+ * @param options	address of popt option table
+ * @param configPaths	colon separated file path(s) to read.
+ * @return		initialized popt context (NULL on error).
+ */
+/*@only@*/ /*@null@*/ /*@unused@*/
+poptContext poptInit(int argc, /*@dependent@*/ /*@keep@*/ const char ** argv,
+		/*@dependent@*/ /*@keep@*/ const struct poptOption * options,
+		/*@null@*/ const char * configPaths)
+	/*@globals fileSystem, internalState @*/
+	/*@modifies fileSystem, internalState @*/;
+
 /** \ingroup popt
  * Reinitialize popt context.
  * @param con		context
@@ -314,15 +352,6 @@ const char ** poptGetArgs(/*@null@*/poptContext con)
 const char * poptBadOption(/*@null@*/poptContext con, unsigned int flags)
 	/*@*/;
 
-/** \ingroup popt
- * Destroy context.
- * @param con		context
- * @return		NULL always
- */
-/*@null@*/
-poptContext poptFreeContext( /*@only@*/ /*@null@*/ poptContext con)
-	/*@modifies con @*/;
-
 /** \ingroup popt
  * Add arguments to context.
  * @param con		context
@@ -356,6 +385,29 @@ int poptAddAlias(poptContext con, struct poptAlias alias, int flags)
 int poptAddItem(poptContext con, poptItem newItem, int flags)
 	/*@modifies con @*/;
 
+/** \ingroup popt
+ * Perform sanity checks on a file path.
+ * @param fn		file name
+ * @return		0 on OK, 1 on NOTOK.
+ */
+int poptSaneFile(const char * fn)
+	/*@globals errno, internalState @*/
+	/*@modifies errno, internalState @*/;
+
+/**
+ * Read a file into a buffer.
+ * @param fn		file name
+ * @retval *bp		buffer (malloc'd) (or NULL)
+ * @retval *nbp		no. of bytes in buffer (including final NUL) (or NULL)
+ * @param flags		1 to trim escaped newlines
+ * return		0 on success
+ */
+int poptReadFile(const char * fn, /*@null@*/ /*@out@*/ char ** bp,
+		/*@null@*/ /*@out@*/ size_t * nbp, int flags)
+	/*@globals errno, fileSystem, internalState @*/
+	/*@modifies *bp, *nbp, errno, fileSystem, internalState @*/;
+#define	POPT_READFILE_TRIMNEWLINES	1
+
 /** \ingroup popt
  * Read configuration file.
  * @param con		context
@@ -367,6 +419,20 @@ int poptReadConfigFile(poptContext con, const char * fn)
 	/*@modifies con->execs, con->numExecs,
 		errno, fileSystem, internalState @*/;
 
+/** \ingroup popt
+ * Read configuration file(s).
+ * Colon separated files to read, looping over poptReadConfigFile().
+ * Note that an '@' character preceeding a path in the list will
+ * also perform additional sanity checks on the file before reading.
+ * @param con		context
+ * @param paths		colon separated file name(s) to read
+ * @return		0 on success, POPT_ERROR_BADCONFIG on failure
+ */
+int poptReadConfigFiles(poptContext con, /*@null@*/ const char * paths)
+	/*@globals errno, fileSystem, internalState @*/
+	/*@modifies con->execs, con->numExecs,
+		errno, fileSystem, internalState @*/;
+
 /** \ingroup popt
  * Read default configuration from /etc/popt and $HOME/.popt.
  * @param con		context
