@@ -42,10 +42,57 @@ typedef struct {
 #define PBM_CLR(d, s)   (__PBM_BITS (s)[__PBM_IX (d)] &= ~__PBM_MASK (d))
 #define PBM_ISSET(d, s) ((__PBM_BITS (s)[__PBM_IX (d)] & __PBM_MASK (d)) != 0)
 
+/** \ingroup popt
+ * Typedef's for string and array of strings.
+ */
+/*@-exporttype@*/
+typedef const char * poptString;
+typedef poptString * poptArgv;
+/*@=exporttype@*/
+
+/** \ingroup popt
+ * A union to simplify opt->arg access without casting.
+ */
+/*@-exporttype -fielduse@*/
+typedef union poptArg_u {
+/*@shared@*/
+    void * ptr;
+    int * intp;
+    long * longp;
+    long long * longlongp;
+    float * floatp;
+    double * doublep;
+    const char ** argv;
+    poptCallbackType cb;
+/*@shared@*/
+    poptOption opt;
+} poptArg;
+/*@=exporttype =fielduse@*/
+
+/*@-exportvar@*/
+/*@unchecked@*/
+extern unsigned int _poptArgMask;
+/*@unchecked@*/
+extern unsigned int _poptGroupMask;
+/*@=exportvar@*/
+
+#define	poptArgType(_opt)	((_opt)->argInfo & _poptArgMask)
+#define	poptGroup(_opt)		((_opt)->argInfo & _poptGroupMask)
+
+#define	F_ISSET(_opt, _FLAG)	((_opt)->argInfo & POPT_ARGFLAG_##_FLAG)
+#define	LF_ISSET(_FLAG)		(argInfo & POPT_ARGFLAG_##_FLAG)
+#define	CBF_ISSET(_opt, _FLAG)	((_opt)->argInfo & POPT_CBFLAG_##_FLAG)
+
+/* XXX sick hack to preserve pretense of a popt-1.x ABI. */
+#define	poptSubstituteHelpI18N(opt) \
+  { /*@-observertrans@*/ \
+    if ((opt) == poptHelpOptions) (opt) = poptHelpOptionsI18N; \
+    /*@=observertrans@*/ }
+
 struct optionStackEntry {
     int argc;
 /*@only@*/ /*@null@*/
-    const char ** argv;
+    poptArgv argv;
 /*@only@*/ /*@null@*/
     pbm_set * argb;
     int next;
@@ -63,7 +110,7 @@ struct poptContext_s {
 /*@dependent@*/
     struct optionStackEntry * os;
 /*@owned@*/ /*@null@*/
-    const char ** leftovers;
+    poptArgv leftovers;
     int numLeftovers;
     int nextLeftover;
 /*@keep@*/
@@ -79,13 +126,14 @@ struct poptContext_s {
     poptItem execs;
     int numExecs;
 /*@only@*/ /*@null@*/
-    const char ** finalArgv;
+    poptArgv finalArgv;
     int finalArgvCount;
     int finalArgvAlloced;
+/*@null@*/
     int (*maincall) (int argc, const char **argv);
 /*@dependent@*/ /*@null@*/
     poptItem doExec;
-/*@only@*/
+/*@only@*/ /*@null@*/
     const char * execPath;
     int execAbsolute;
 /*@only@*/ /*@relnull@*/
@@ -94,26 +142,9 @@ struct poptContext_s {
     pbm_set * arg_strip;
 };
 
-#ifdef HAVE_LIBINTL_H
-#include <libintl.h>
-#endif
-
-#if defined(HAVE_GETTEXT) && !defined(__LCLINT__)
-#define _(foo) gettext(foo)
+#if defined(POPT_fprintf)
+#define	POPT_dgettext	dgettext
 #else
-#define _(foo) foo
-#endif
-
-#if defined(HAVE_DCGETTEXT) && !defined(__LCLINT__)
-#define D_(dom, str) dgettext(dom, str)
-#define POPT_(foo) D_("popt", foo)
-#else
-#define D_(dom, str) str
-#define POPT_(foo) foo
-#endif
-
-#define N_(foo) foo
-
 #ifdef HAVE_ICONV
 #include <iconv.h>
 #if defined(__LCLINT__)
@@ -122,9 +153,9 @@ extern /*@only@*/ iconv_t iconv_open(const char *__tocode, const char *__fromcod
 	/*@*/;
 
 extern size_t iconv(iconv_t __cd, /*@null@*/ char ** __inbuf,
-		    /*@out@*/ size_t * __inbytesleft,
-		    /*@out@*/ char ** __outbuf,
-		    /*@out@*/ size_t * __outbytesleft)
+		    /*@null@*/ /*@out@*/ size_t * __inbytesleft,
+		    /*@null@*/ /*@out@*/ char ** __outbuf,
+		    /*@null@*/ /*@out@*/ size_t * __outbytesleft)
 	/*@modifies __cd,
 		*__inbuf, *__inbytesleft, *__outbuf, *__outbytesleft @*/;
 
@@ -144,11 +175,41 @@ extern char *nl_langinfo (nl_item __item)
 #endif
 #endif
 
+#if defined(HAVE_DCGETTEXT) && !defined(__LCLINT__)
+char *POPT_dgettext(const char * dom, const char * str)
+	/*@*/;
+#endif
+
 int   POPT_fprintf (FILE* stream, const char *format, ...)
 	/*@globals fileSystem @*/
 	/*@modifies stream, fileSystem @*/;
+#endif	/* !defined(POPT_fprintf) */
 
-char *POPT_prev_char (/*@returned@*/ const char *str)
+const char *POPT_prev_char (/*@returned@*/ const char *str)
+	/*@*/;
+
+const char *POPT_next_char (/*@returned@*/ const char *str)
 	/*@*/;
 
 #endif
+
+#if defined(ENABLE_NLS) && defined(HAVE_LIBINTL_H)
+#include <libintl.h>
+#endif
+
+#if defined(ENABLE_NLS) && defined(HAVE_GETTEXT) && !defined(__LCLINT__)
+#define _(foo) gettext(foo)
+#else
+#define _(foo) foo
+#endif
+
+#if defined(ENABLE_NLS) && defined(HAVE_DCGETTEXT) && !defined(__LCLINT__)
+#define D_(dom, str) POPT_dgettext(dom, str)
+#define POPT_(foo) D_("popt", foo)
+#else
+#define D_(dom, str) str
+#define POPT_(foo) foo
+#endif
+
+#define N_(foo) foo
+
